/*
   Hangul Automata

   Written by Song Jaekyung
 */

#include "hangul.h"

#define	F_F	0x1				/* ʼ */
#define F_M	0x2				/* ߼ */
#define F_L	0x4				/*  */
#define F_A	0x8				/* ASCII */
#define F_NULL 1
#define M_NULL 2
#define L_NULL 1
enum { Nothing,F_ing,M_ing,L_ing} ing;

int keyboard = 3;
int hangul_state = 0;
int usehangul = 0;


static int f=F_NULL, m=M_NULL, l=L_NULL;
static int sp=0;
static unsigned stack[20];
static char c_stack1[20],c_stack2[20];
static char f_stack[20],m_stack[20],l_stack[20],ing_stack[20];


#define push(x)	( stack[ sp++ ] = *(x);stack[sp++] = *((x)+1) )
#define pop()	( stack[ --sp ] )

#define push3(buf) { f_stack[sp]=f;m_stack[sp]=m;l_stack[sp]=l;\
         ing_stack[sp]=ing;c_stack1[sp]=*(buf);c_stack2[sp++]=*((buf)+1);}
#define pop3(buf) { sp-=2;f=f_stack[sp];m=m_stack[sp];l=l_stack[sp];\
		 ing=ing_stack[sp];*(buf)=c_stack1[sp];*((buf)+1)=c_stack2[sp++];}
/* table_for_3   flag */


int (*convert_3_to_code)(int f,int m,int l,unsigned char *buf);

/* 3 Ŀ  ȯ */

static unsigned char value_table_for_3[] =
{
	24  , '"' , '#' , '$' , '%' , '&' ,	/* ! " # $ % & */
	18  , '(' , ')' , '*' , '+' , ',' ,	/* ' ( ) * + , */
	'-' , '.' , 13  , 17  , 29  , 22  ,	/* - . / 0 1 2 */
	19  , 19  , 26  , 5   , 12  , 28  ,	/* 3 4 5 6 7 8 */
	20  , ':' , 9   , '2' , '=' , '3' ,	/* 9 : ; < = > */
	'?' , '@' , 8   , '!' , 11  , 10  ,	/* ? @ A B C D */
	26  , 3   , '/' , 39  , '8' , '4' ,	/* E F G H I J */
	'5' , '6' , '1' , '0' , '9' , '>' ,	/* K L M N O P */
	28  , 6   , 7   , ';' , '7' , 16  ,	/* Q R S T U V */
	27  , 20  , '<' , 25  , '[' , 92  ,	/* W X Y Z [ \ */
	']' , '^' , '_' , '`' , 23  , 20  ,	/* ] ^ _ ` a b */
	10  , 29  , 11  , 3   , 27  , 4   ,	/* c d e f g h */
	8   , 13  , 2   , 14  , 20  , 11  ,	/* i j k l m n */
	16  , 19  , 21  , 4   , 5   , 7   ,	/* o p q r s t */
	5   , 13  , 9   , 2   , 7   , 17  ,	/* u v w x y z */
};

static unsigned short kind_table_for_3[] =
{
	F_L, F_A, F_A, F_A, F_A, F_A,	/* ! " # $ % & */
	F_F, F_A, F_A, F_A, F_A, F_A,	/* ' ( ) * + , */
	F_A, F_A, F_M, F_F, F_L, F_L,	/* - . / 0 1 2 */
	F_L, F_M, F_M, F_M, F_M, F_M,	/* 3 4 5 6 7 8 */
	F_M, F_A, F_F, F_A, F_A, F_A,	/* 9 : ; < = > */
	F_A, F_A, F_L, F_A, F_L, F_L,	/* ? @ A B C D */
	F_L, F_L, F_A, F_A, F_A, F_A,	/* E F G H I J */
	F_A, F_A, F_A, F_A, F_A, F_A,	/* K L M N O P */
	F_L, F_M, F_L, F_A, F_A, F_L,	/* Q R S T U V */
	F_L, F_L, F_A, F_L, F_A, F_A,	/* W X Y Z [ \ */
	F_A, F_A, F_A, F_A, F_L, F_M,	/* ] ^ _ ` a b */
	F_M, F_M, F_M, F_M, F_M, F_F,	/* c d e f g h */
	F_F, F_F, F_F, F_F, F_F, F_F,	/* i j k l m n */
	F_F, F_F, F_L, F_M, F_L, F_M,	/* o p q r s t */
	F_F, F_M, F_L, F_L, F_F, F_L,	/* u v w x y z */
};

/* 3 Ŀ (ʼ, Է¿) -> ʼ ó */

static int 
comfcon3(v, c)
int v, c;
{
	if (v == 2 && c == 2)
		return 3;
	if (v == 5 && c == 5)
		return 6;
	if (v == 9 && c == 9)
		return 10;
	if (v == 11 && c == 11)
		return 12;
	if (v == 14 && c == 14)
		return 15;
	return 0;
}

/* 3 Ŀ (, Է ) ->   ó */

static int 
comvow3(v, c)
int v, c;
{
	switch (v) {
	case 13:					/*  */
		switch (c) {
		case 3:				/* Ǥ */
			return 14;
		case 4:				/* Ǥ */
			return 15;
		case 29:				/* Ǥ */
			return 18;
		}
		break;
	case 20:					/*  */
		switch (c) {
		case 7:				/* ̤ */
			return 21;
		case 10:				/* ̤ */
			return 22;
		case 29:				/* ̤ */
			return 23;
		}
		break;

		/* 3   Ѥ  Ƿ ... */
	}
	return 0;
}

/* 3 Ŀ ( ħ,  Է) -> ħ */

static int 
comcon3(int k, int c)
{
	switch (k) {
	case 2:					/*  */
		switch (c) {
		case 2:
			return 3;			/*  */
		case 21:
			return 4;			/*  */
		}
		break;
	case 5:					/*  */
		switch (c) {
		case 24:				/*  */
			return 6;
		case 29:
			return 7;			/*  */
		}
		break;
	case 9:					/*  */
		switch (c) {
		case 2:				/*  */
			return 10;
		case 17:				/*  */
			return 11;
		case '3':				/*  */
			return 12;
		case 21:				/*  */
			return 13;
		case 27:				/*  */
			return 14;
		case 28:				/*  */
			return 15;
		case 29:				/*  */
			return 16;
		}
		break;
	case 19:
		switch (c) {
		case 21:				/*  */
			return 20;
		}
		break;
	}
	return 0;
}

/* 3 Ŀ  ó */

static int 
ascii3(c, buf)
int c;
char *buf;
{
	int rs;
	if (f > 1 || m > 2 || l > 1) {
		rs = convert_3_to_code (f, m, l, buf);
		sp = 0;
		f = 1; m = 2; l = 1;
	} else {
		rs = 0;
	}
	*(buf+rs) = c;
	return rs + 1;
}

/* 2 Ŀ  ->  ʼ ȯ
   : ʼ ƴϸ 0
 */
static int fcon(c)
int c;
{
	static char table[] =
	{
	/*E */ 6, /*F */ 0, /*G */ 0, /*H */ 0, /*I */ 0, /*J */ 0, /*K */ 0,
	/*L */ 0, /*M */ 0, /*N */ 0, /*O */ 0, /*P */ 0, /*Q */ 10, /*R */ 3,
		/*S */ 0, /*T */ 12, /*U */ 0, /*V */ 0, /*W */ 15, /*X */ 0, /*Y */ 0,
	 /*Z */ 0, /*[ */ 0, /*\ */ 0, /*] */ 0, /*^ */ 0, /*_*/ 0, /*` */ 0,
		/*a */ 8, /*b */ 0, /*c */ 16, /*d */ 13, /*e */ 5, /*f */ 7, /*g */ 20,
	/*h */ 0, /*i */ 0, /*j */ 0, /*k */ 0, /*l */ 0, /*m */ 0, /*n */ 0,
	/*o */ 0, /*p */ 0, /*q */ 9, /*r */ 2, /*s */ 4, /*t */ 11, /*u */ 0,
		/*v */ 19, /*w */ 14, /*x */ 18, /*y */ 0, /*z */ 17};

	if (c < 'E' || c > 'z')
		return 0;
	return table[c - 'E'];
}

/* 2 Ŀ  -> ߼ ȯ
   : ߼ ƴϸ 0
 */
static int vow(c)
int c;
{
	static char table[] =
	{
	/*O */ 6, /*P */ 12, /*Q */ 0, /*R */ 0, /*S */ 0, /*T */ 0, /*U */ 0,
	/*V */ 0, /*W */ 0, /*X */ 0, /*Y */ 0, /*Z */ 0, /*[ */ 0, /*\ */ 0,
	/*] */ 0, /*^ */ 0, /*_*/ 0, /*` */ 0, /*a */ 0, /*b */ 26, /*c */ 0,
	/*d */ 0, /*e */ 0, /*f */ 0, /*g */ 0, /*h */ 13, /*i */ 5, /*j */ 7,
		/*k */ 3, /*l */ 29, /*m */ 27, /*n */ 20, /*o */ 4, /*p */ 10, /*q */ 0,
	/*r */ 0, /*s */ 0, /*t */ 0, /*u */ 11, /*v */ 0, /*w */ 0, /*x */ 0,
	/*y */ 19};

	if (c < 'O' || c > 'y')
		return 0;
	return table[c - 'O'];
}

/* 2Ŀ  -> ħ ȯ
   : ħ ƴϸ 0
 */
static int lcon(c)
int c;
{
	static char table[] =
	{
	/*R */ 3, /*S */ 0, /*T */ 22, /*U */ 0, /*V */ 0, /*W */ 0, /*X */ 0,
	 /*Y */ 0, /*Z */ 0, /*[ */ 0, /*\ */ 0, /*] */ 0, /*^ */ 0, /*_*/ 0,
		/*` */ 0, /*a */ 17, /*b */ 0, /*c */ 25, /*d */ 23, /*e */ 8, /*f */ 9,
	/*g */ 29, /*h */ 0, /*i */ 0, /*j */ 0, /*k */ 0, /*l */ 0, /*m */ 0,
		/*n */ 0, /*o */ 0, /*p */ 0, /*q */ 19, /*r */ 2, /*s */ 5, /*t */ 21,
		/*u */ 0, /*v */ 28, /*w */ 24, /*x */ 27, /*y */ 0, /*z */ 26};

	if (c < 'R' || c > 'z')
		return 0;
	return table[c - 'R'];
}

/* 2 Ŀ ( ħ,  Է) -> ħ ȯ */

static int comcon(k, c)
int k;
int c;
{
	switch (k) {
	case 2:					/*  */
		switch (c) {
		case 't':
			return 4;			/*  */
		}
		break;
	case 5:					/*  */
		switch (c) {
		case 'w':				/*  */
			return 6;
		case 'g':				/*  */
			return 7;
		}
		break;
	case 9:					/*  */
		switch (c) {
		case 'r':				/*  */
			return 10;
		case 'a':				/*  */
			return 11;
		case 'q':				/*  */
			return 12;
		case 't':				/*  */
			return 13;
		case 'x':				/*  */
			return 14;
		case 'v':				/*  */
			return 15;
		case 'g':				/*  */
			return 16;
		}
		break;
	case 19:					/*  */
		switch (c) {
		case 't':				/*  */
			return 20;
		}
		break;
	}
	return 0;
}

/* 2Ŀ ( ߼,  Է) -> ߼ ȯ */

static int comvow(v, c)
int v;
int c;
{
	switch (v) {
	case 13:					/*  */
		switch (c) {
		case 'k':				/* Ǥ */
			return 14;
		case 'o':				/* Ǥ */
			return 15;
		case 'l':				/* Ǥ */
			return 18;
		}
		break;
	case 20:					/*  */
		switch (c) {
		case 'j':				/* ̤ */
			return 21;
		case 'p':				/* ̤ */
			return 22;
		case 'l':				/* ̤ */
			return 23;
		}
		break;
	case 27:					/*  */
		switch (c) {
		case 'l':				/* Ѥ */
			return 28;
		}
		break;
	}
	return 0;
}

/* 2 Ŀ  ó */

int
ascii(c, buf)
int c;
char *buf;
{
	if (c == -1) {
		if (hangul_state)
			hangul_state = 0;
		else {
			hangul_state = 1;
			f = 1;
			m = 2;
			l = 1;
		}
		return 0;
	} else if (c == -2) {		/* flush output */
		f = 1;
		m = 2;
		l = 1;
		return 0;
	} else {
		buf[0] = c;
		return 1;
	}
}

/* ( Է) -> ϼ ڵ ڿ
   c :  ASCII
   buf : ϼ  KS C 5601 - 1992 ڵ
   : buf   Ʈ  °
   term->screen.keyboard :  . 2 = 2, 3 = 3
   convert_3_to_ks() : ->ϼ ȯ. ȯ  ڿ ̸
     Ѵ. 2 Ǵ 8
   show_status() : ѱ     ǥѴ.
 */

int (*hangul_automata)(int c,char *buf);

int 
hangul_automata2(c, buf)
int c;							/* input character */
char *buf;
{
	int t, t2, rs;
	return 1;
}


/**/
int 
hangul_automata3(int c,char *buf)
{
	int t, t2;
	int ret,ret2;
	int beepifnotexist=1;
	if( c >= '!' && c <= 'z' ) {
		c -= '!';
		t = value_table_for_3[c];
		switch(kind_table_for_3[c]) {
			case F_F: /* ʼ */
			    if( f == F_NULL) { /* ʼ  */
				  ing=F_ing;
				  f = t;
				  ret = convert_3_to_code(f,m,l,buf);
				  push3(buf);
			 	  if(sp == 1) return -ret;
				  else return ret;
			    }
			    if(ing == F_ing) { /* ʼ   */
					 if(t2=comfcon3(f,t)) { /*  */
						 f=t2;
						 ing=Nothing;
						 ret = convert_3_to_code(f,m,l,buf);
						 push3(buf);
				         return ret;
					  }
			    }
			    ret = convert_3_to_code(f,m,l,buf);
			    f=t;m=M_NULL;l=L_NULL;
			    ret2 = convert_3_to_code(f,m,l,buf+ret);
			    ing=F_ing;
			    sp=0;
			    push3(buf+ret);
			    return ret+ret2;
			case F_M: /*  */
				if( m == M_NULL) {
					ing = M_ing;
					m = t;
					ret = convert_3_to_code(f,m,l,buf);
					push3(buf);
					if(sp == 1) return -ret;
					else return ret;
				}
				if(ing == M_ing) {
					if(t2 = comvow3(m,t)) { /*  */
						ing=Nothing;
						m = t2;
						ret = convert_3_to_code(f,m,l,buf);
						push3(buf);
						return ret;
					}
				}
				ret = convert_3_to_code(f,m,l,buf);
				m=t;f=F_NULL;l=L_NULL;
				ret2 = convert_3_to_code(f,m,l,buf+ret);
				sp=0;
				push3(buf+ret);
				ing=M_ing;
				return ret+ret2;
			case F_L: /* ħ */
				if( l == L_NULL) {
					ing = L_ing;
					l = t;
					ret = convert_3_to_code(f,m,l,buf);
					if(ret != 2 && beepifnotexist) {
						l = L_NULL;
						return 10;
					}
					push3(buf);
					if(sp == 1) return -ret;
					else return ret;
				}
				if(ing == L_ing) {
					if(t2 = comcon3(l,t)) {
						ing = Nothing;
						l = t2;
						ret = convert_3_to_code(f,m,l,buf);
						if(ret != 2 && beepifnotexist) {
							l = l_stack[sp-1];
							return 10;
						}
						push3(buf);
						return ret;
					}
				}
				ret = convert_3_to_code(f,m,l,buf);
				if(ret != 2 && beepifnotexist) {
					return 10;
				}
				l=t;f=F_NULL;m=M_NULL;
				ret2 = convert_3_to_code(f,m,l,buf+ret);
				if(ret2 != 2 && beepifnotexist) {
					f = f_stack[sp-1];
					l = l_stack[sp-1];
					m = m_stack[sp-1];
					return 10;
				}
				sp=0;
				push3(buf+ret);
				ing=L_ing;
				return ret+ret2;
			case F_A: /* Ưڳ  */

				return ascii3(t,buf);

		} /* end of switch */
	} else  if (c == '\b' || c == 0x7f ) {
		if (sp > 0) {
			pop3(buf);			/*  ڴ  */
			if (sp > 0) {
				sp--;
				f = f_stack[sp];
				m = m_stack[sp];
				l = l_stack[sp];
				ing = ing_stack[sp];
				*(buf) = c_stack1[sp];
				*(buf+1) = c_stack2[sp];
				sp++;
				return 2;
			}
			f = 1; m = 2; l = 1;
			return 0; /* stack is now empty and no return char */
		} else {
			return -1;
		}
	}
	if(sp==0) return -3; /* stack is already empty and c is special character */
	sp=0; f=F_NULL;m=M_NULL;l=L_NULL;
	return -30; /* stack is now empty, c is special */
}

/**/
int
in_ks(f, m, t)
int f, m, t;
{
	char buf[8];

	return convert_3_to_ks(f, m, t, buf) == 2;
}

#undef F_NULL
#undef M_NULL
#undef L_NULL
